﻿using Kivii.Finances.Entities;
using Kivii.Linq;
using Kivii.Web;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Finances.Transforms
{
    #region 申请开票相关
    [Api(Description = "开票申请")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyApply : RestfulExecution<InvoiceApply>
    {
        public InvoiceApply Item { get; set; }

        public List<Guid> PaymentKvids { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            Item.ThrowIfNull("请传入要申请的内容!");
            if (Item.PayerName.IsNullOrEmpty()) throw new Exception("请输入发票抬头!");
            if (Item.Currency == CurrencyUnit.Unsupported) throw new Exception("不支持的货币单位!");
            if (Item.Amount <= 0) throw new Exception("申请金额不能小于0!");
            if (Item.OperateType != InvoiceApplyType.Debit) PaymentKvids.ThrowIfNullOrEmpty("请先选择相应收款!");

            //申请人为当前登录人
            Item.OperatorName = KiviiContext.CurrentMember.FullName;
            Item.OperatorKvid = KiviiContext.CurrentMember.Kvid;
            //申请所属部门
            Item.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
            Item.OwnerName = KiviiContext.CurrentMember.DepartmentName;

            var rtns = new RestfulQueryResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();

            if (Item.OperateType == InvoiceApplyType.Debit)
            {
                var applyKvid = Guid.NewGuid();
                Item.RootKvid = applyKvid;
                Item.ParentKvid = Guid.Empty;
                //Item.AmountUsed = 0;
                Item.OperateTime = DateTime.Now;
                Item.Status = (int)InvoiceApplyStatus.FinancialApproval;
                Item.Kvid = applyKvid;
                conn.Insert(Item);
                rtns.Results.Add(Item);
            }
            else
            {
                var paymentKvidsDistinct = PaymentKvids.Distinct().ToList();
                var payments = conn.Select<Payment>(o => Sql.In(o.Kvid, paymentKvidsDistinct));
                (payments.Count <= 0 || payments.Count != paymentKvidsDistinct.Count).ThrowIfTrue("收款信息与所选项目不一致!");
                (payments.Exists(o => o.Type != PaymentType.Bank && o.Type != PaymentType.Split && o.Type != PaymentType.AliPay && o.Type != PaymentType.Pos && o.Type != PaymentType.WeChat && o.Type != PaymentType.Cash)).ThrowIfTrue("请选择正确的收款信息进行申请!");
                (payments.Exists(o => o.Amount <= o.AmountInvoice)).ThrowIfTrue("存在已开票的到账!无法重复申请!");
                if (Item.Amount != payments.Sum(o => o.Amount - o.AmountInvoice)) throw new Exception("申请金额与所选到账金额不一致!");

                (conn.Exists<InvoiceApply>(o => o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty && Sql.In(o.BizKvid, payments.ConvertAll(p => p.Kvid)))).ThrowIfTrue("已有申请开票中的到账!");
                //(conn.Exists<Invoice>(o => o.OffsetKvid == Guid.Empty && Sql.In(o.BizKvid, payments.ConvertAll(p => p.RootKvid)))).ThrowIfTrue("已有开票的到账!");

                var trans = conn.OpenTransaction();
                try
                {
                    var applyKvid = Guid.NewGuid();
                    Item.RootKvid = applyKvid;
                    Item.ParentKvid = Guid.Empty;
                    //Item.AmountUsed = Item.Amount;
                    Item.OperateTime = DateTime.Now;
                    Item.Status = (int)InvoiceApplyStatus.FinancialApproval;
                    Item.Kvid = applyKvid;
                    conn.Insert(Item);
                    rtns.Results.Add(Item);

                    foreach (var item in payments)
                    {
                        var relation = new InvoiceApply();
                        relation.PopulateWith(Item);
                        relation.BizId = item.SerialNumber;
                        relation.BizKvid = item.Kvid;
                        relation.BizType = typeof(Payment).FullName;
                        relation.Kvid = Guid.NewGuid();
                        relation.RootKvid = applyKvid;
                        relation.ParentKvid = applyKvid;
                        relation.SerialNumber = string.Empty;
                        relation.OperateType = InvoiceApplyType.Related;
                        relation.Type = Item.OperateType.ToString();
                        relation.Amount = item.Amount - item.AmountInvoice;
                        relation.OperatorName = KiviiContext.CurrentMember.FullName;
                        relation.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                        relation.AmountInvoice = 0;
                        //relation.AmountUsed = relation.Amount;
                        conn.Insert(relation);
                    }

                    trans.Commit();
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    throw ex;
                }
            }

            rtns.Total = rtns.Results.Count();
            return rtns;
        }
    }

    [Api(Description = "批量借票申请")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyMultiApply : RestfulExecution<InvoiceApply>
    {
        public InvoiceApply Item { get; set; }

        public List<InvoiceApply> Items { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Item == null && Items.IsNullOrEmpty()).ThrowIfTrue("请传入要借票信息!");
            if (Items.IsNullOrEmpty()) Items = new List<InvoiceApply>();
            if (Item != null) Items.Add(Item);
            if (Items.Exists(o => o.PayerName.IsNullOrEmpty())) throw new Exception("存在发票抬头为空!");
            if (Items.Exists(o => o.PayerTaxNumber.IsNullOrEmpty())) throw new Exception("存在税号为空!");
            if (Items.Exists(o => o.Currency == CurrencyUnit.Unsupported)) throw new Exception("存在不支持的货币单位!");
            if (Items.Exists(o => o.Amount <= 0)) throw new Exception("申请金额不能小于0!");
            if (Items.Exists(o => o.OperateType != InvoiceApplyType.Debit)) throw new Exception("此接口仅针对借票申请开票!");

            var rtns = new RestfulQueryResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();

            foreach (var item in Items)
            {
                #region 创建借票申请
                var applyKvid = Guid.NewGuid();
                item.OperateType = InvoiceApplyType.Debit;
                //申请人为当前登录人
                item.OperatorName = KiviiContext.CurrentMember.FullName;
                item.OperatorKvid = KiviiContext.CurrentMember.Kvid;
                //申请所属部门
                item.OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                item.OwnerName = KiviiContext.CurrentMember.DepartmentName;
                item.RootKvid = applyKvid;
                item.ParentKvid = Guid.Empty;
                //item.AmountUsed = 0;
                item.OperateTime = DateTime.Now;
                item.Status = (int)InvoiceApplyStatus.FinancialApproval;
                item.Kvid = applyKvid;
                conn.Insert(item);
                #endregion
                rtns.Results.Add(item);
            }

            rtns.Total = rtns.Results.Count();
            return rtns;
        }
    }
    #endregion

    #region 申请处理相关
    [Api(Description = "申请驳回")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyReject : RestfulExecution<InvoiceApply>
    {
        public List<Guid> ApplyKvids { get; set; }

        public string Summary { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var applys = conn.Select<InvoiceApply>(o => Sql.In(o.Kvid, ApplyKvids) && o.OperateType != InvoiceApplyType.Related);
            if (applys == null) throw new Exception("无相关申请!");
            var applyRelations = conn.Select<InvoiceApply>(o => Sql.In(o.ParentKvid, applys.ConvertAll(p => p.Kvid)) && o.OperateType == InvoiceApplyType.Related);

            var rtns = new RestfulExecutionResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var trans = conn.OpenTransaction();
            try
            {
                foreach (var apply in applys)
                {
                    #region apply驳回
                    //进入开票状态后不可驳回
                    (apply.Status >= (int)InvoiceApplyStatus.FinancialExecute).ThrowIfTrue("撤销失败,当前申请已进入开票列队或已完成开票!");
                    apply.Status = (int)InvoiceApplyStatus.ApplyReject;
                    apply.AddOnlyProperties(o => o.Status);
                    //apply.AmountUsed = 0;
                    //apply.AddOnlyProperties(o => o.AmountUsed);
                    apply.Summary += "  (驳回原因:" + Summary + ")";
                    apply.AddOnlyProperties(o => o.Summary);
                    conn.UpdateOnly(apply);
                    apply.RemoveAllOnlyProperties();
                    rtns.Results.Add(apply);

                    if (applyRelations.IsNullOrEmpty()) continue;
                    var relations = applyRelations.Where(o => o.ParentKvid == apply.Kvid).ToList();
                    if (relations.IsNullOrEmpty()) continue;
                    #endregion

                    #region 关联apply 删除
                    foreach (var item in relations)
                    {
                        item.Status = -1;
                        item.AddOnlyProperties(o => o.Status);
                        conn.UpdateOnly(item);
                    }
                    #endregion
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return rtns;
        }
    }

    [Api(Description = "执行开票")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyExecute : RestfulExecution<InvoiceApply>
    {
        public InvoiceApply Item { get; set; }

        public InvoiceApplyDetail Detail { get; set; }

        public List<InvoiceApplyDetail> Details { get; set; }

        //发票开具最大额度限制 默认10万
        public decimal Limit { get; set; } = 100000;

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Detail == null && Details.IsNullOrEmpty()).ThrowIfTrue("Need Detail Or Details!");
            if (Details.IsNullOrEmpty()) Details = new List<InvoiceApplyDetail>();
            if (Detail != null) Details.Add(Detail);

            if (Item.Kvid == Guid.Empty) throw new Exception("请先选择申请条目!");
            if (Details.Exists(o => o.GoodsFullName.IsNullOrEmpty())) throw new Exception("明细名称存在空值!");
            if (Details.Exists(o => o.TaxRate < 0) || Details.Exists(o => o.TaxRate > 1)) throw new Exception("存在明细税率设置范围不在0-1之间!");
            if (Details.Exists(o => o.Quantity <= 0)) throw new Exception("请填写数量!");
            if (Details.Count > 1)
            {
                var group = Details.GroupBy(o => o.GoodsId);
                foreach (var kv in group)
                {
                    var sum = kv.Sum(o => o.Amount);
                    if (sum > Limit) throw new Exception($"组别:{kv.Key},总额:{sum}元,金额总和不能超过限额{Limit}元!");
                }
            }

            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var apply = conn.SingleById<InvoiceApply>(Item.Kvid);
            if (apply == null) throw new Exception("不存在的申请!");
            (apply.OperateType == InvoiceApplyType.Related).ThrowIfTrue("申请类型不正确!");
            if (apply.Status >= (int)InvoiceApplyStatus.FinancialExecute) throw new Exception("请勿重复申请开票!");
            if (apply.Status < (int)InvoiceApplyStatus.ProcessAdoption) throw new Exception("此申请还在审批中...");
            if (apply.Amount != Details.Sum(o => o.Amount)) throw new Exception("总金额和明细总额不一致!");
            conn.InitEntityType<InvoiceApplyDetail>();
            var existDetails = conn.Select<InvoiceApplyDetail>(o => o.ApplyKvid == apply.Kvid);
            var applyRelations = conn.Select<InvoiceApply>(o => o.ParentKvid == apply.Kvid && o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty);
            var payments = conn.Select<Payment>(o => Sql.In(o.Kvid, applyRelations.ConvertAll(p => p.BizKvid)));//Type:Split,Pos,Cash,WeChat...
            (conn.Exists<Invoice>(o => o.OffsetKvid != Guid.Empty && Sql.In(o.BizKvid, payments.ConvertAll(p => p.RootKvid)))).ThrowIfTrue("已有开票的到账,无法重复开票,请驳回!");

            #region 计算(不根据前端传来)
            foreach (var detail in Details)
            {
                detail.AmountUntaxed = Math.Round(detail.Amount / (1 + detail.TaxRate), 2);
                detail.AmountTax = detail.Amount - detail.AmountUntaxed;
                detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                detail.QuantityUnitPrice = Math.Round(detail.Amount / detail.Quantity, 2);
            }
            #endregion

            var rtns = new RestfulExecutionResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var trans = conn.OpenTransaction();
            try
            {
                #region 申请状态进入开票状态
                apply.Status = (int)InvoiceApplyStatus.FinancialExecute;
                apply.AddOnlyProperties(o => o.Status);
                apply.Summary = Item.Summary;
                apply.AddOnlyProperties(o => o.Summary);
                apply.Remark = Item.Remark;
                apply.AddOnlyProperties(o => o.Remark);
                apply.Metadata["PayeeOperatorName"] = Item.Metadata.ContainsKey("PayeeOperatorName") ? Item.Metadata["PayeeOperatorName"] : KiviiContext.CurrentMember.FullName;
                apply.Metadata["ReviewerName"] = Item.Metadata.ContainsKey("ReviewerName") ? Item.Metadata["ReviewerName"] : KiviiContext.CurrentMember.FullName;
                apply.AddOnlyProperties(o => o.Metadata);
                conn.UpdateOnly(apply);
                rtns.Results.Add(apply);
                #endregion

                #region 申请明细入库(发票明细)
                foreach (var detail in Details)
                {
                    detail.ApplyKvid = apply.Kvid;
                    conn.Insert(detail);
                }
                #endregion
                if (!existDetails.IsNullOrEmpty())
                {
                    foreach (var exist in existDetails)
                    {
                        exist.Status = -1;
                        exist.AddOnlyProperties(o => o.Status);
                        conn.UpdateOnly(exist);
                    }
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }

            return rtns;
        }
    }

    [Api(Description = "申请合并")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyCombine : RestfulExecution<InvoiceApply>
    {
        public InvoiceApply Item { get; set; }

        public List<Guid> Kvids { get; set; }

        public InvoiceApplyDetail Detail { get; set; }

        public List<InvoiceApplyDetail> Details { get; set; }

        //发票开具最大额度限制 默认10万
        public decimal Limit { get; set; } = 100000;

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Detail == null && Details.IsNullOrEmpty()).ThrowIfTrue("Need Detail Or Details!");
            if (Details.IsNullOrEmpty()) Details = new List<InvoiceApplyDetail>();
            if (Detail != null) Details.Add(Detail);
            if (Kvids == null || Kvids.Count <= 0) throw new Exception("请先选择申请条目!");
            if (Item.OperateType != InvoiceApplyType.Payment) throw new Exception("合并开票只针对收款刷卡现金申请!");

            if (Details.Exists(o => o.GoodsFullName.IsNullOrEmpty())) throw new Exception("明细名称存在空值!");
            if (Details.Exists(o => o.TaxRate < 0) || Details.Exists(o => o.TaxRate > 1)) throw new Exception("存在明细税率设置范围不在0-1之间!");
            if (Details.Exists(o => o.Quantity <= 0)) throw new Exception("请填写数量!");
            var group = Details.GroupBy(o => o.GoodsId);
            foreach (var kv in group)
            {
                var sum = kv.Sum(o => o.Amount);
                if (sum > Limit) throw new Exception($"组别:{kv.Key},总额:{sum}元,金额总和不能超过限额{Limit}元!");
                //if (sum > Limit) throw new Exception($"组别:{kv.Key},总额:{sum}元,金额总和不能超过10万!");
            }
            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var applys = conn.Select<InvoiceApply>(o => Sql.In(o.Kvid, Kvids));

            if (applys.Count <= 0 || applys == null) throw new Exception("不存在的申请!");
            if (applys.Count < 2) throw new Exception("请选择两条以上的申请进行合并!");
            //if (applys.Exists(o => o.Type != "VATS")) throw new Exception("普票暂时无法合并开票");
            if (applys.Exists(o => o.PayerName != Item.PayerName)) throw new Exception("合并申请抬头不一致!");
            if (!applys.Exists(o => o.OperateType != InvoiceApplyType.Payment)) throw new Exception("合并开票只针对收款刷卡现金申请!");
            if (applys.Exists(o => o.OperateType == InvoiceApplyType.Payment)) Item.OperateType = InvoiceApplyType.Payment;
            var groups = applys.GroupBy(o => new { o.PayerName, o.PayerTaxNumber });
            if (groups.Count() != 1) throw new Exception("合并申请抬头不一致!");
            if (applys.Exists(o => o.Status >= (int)InvoiceApplyStatus.FinancialExecute)) throw new Exception("请勿重复申请开票!");
            if (applys.Exists(o => o.Status < (int)InvoiceApplyStatus.ProcessAdoption)) throw new Exception("此申请还在审批中...");
            if (applys.Sum(o => o.Amount) != Item.Amount) throw new Exception("合并金额与申请金额不一致!");
            if (Item.Amount != Details.Sum(o => o.Amount)) throw new Exception("总金额和明细总额不一致!");

            var remark = $"[总: {Item.Amount}元]";
            foreach (var item in applys)
            {
                remark += $"[{item.OwnerName}: {item.Amount}元]";
            }
            #region 计算(不根据前端传来)
            foreach (var detail in Details)
            {
                detail.AmountUntaxed = Math.Round(detail.Amount / (1 + detail.TaxRate), 2);
                detail.AmountTax = detail.Amount - detail.AmountUntaxed;
                detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                detail.QuantityUnitPrice = Math.Round(detail.Amount / detail.Quantity, 2);
            }
            #endregion
            var relationApplys = conn.Select<InvoiceApply>(o => Sql.In(o.ParentKvid, applys.Select(p => p.Kvid)) && o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty);

            var rtns = new RestfulExecutionResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var trans = conn.OpenTransaction();
            try
            {
                var newInvoiceApplyKvid = Guid.NewGuid();
                #region 创建新的Apply
                var newApply = new InvoiceApply();
                newApply.PopulateWith(Item);
                newApply.SerialNumber = string.Empty;
                newApply.RootKvid = newInvoiceApplyKvid;
                newApply.ParentKvid = Guid.Empty;
                newApply.Kvid = newInvoiceApplyKvid;
                newApply.Amount = applys.Sum(o => o.Amount);
                //newApply.AmountUsed = applys.Sum(o => o.AmountUsed);
                newApply.Status = (int)InvoiceApplyStatus.FinancialApproval;
                newApply.Summary = Item.Summary;
                newApply.Remark = remark;
                newApply.Metadata["PayeeOperatorName"] = Item.Metadata.ContainsKey("PayeeOperatorName") ? Item.Metadata["PayeeOperatorName"] : KiviiContext.CurrentMember.FullName;
                newApply.Metadata["ReviewerName"] = Item.Metadata.ContainsKey("ReviewerName") ? Item.Metadata["ReviewerName"] : KiviiContext.CurrentMember.FullName;
                conn.Insert(newApply);
                rtns.Results.Add(newApply);
                #endregion

                #region ApplyDetail入库
                foreach (var detail in Details)
                {
                    detail.ApplyKvid = newInvoiceApplyKvid;
                    conn.Insert(detail);
                }
                #endregion

                #region 老Apply更新(更改状态和biz记录新生产的Apply)
                foreach (var item in applys)
                {
                    //item.AmountUsed = 0;
                    //item.AddOnlyProperties(o => o.AmountUsed);
                    item.BizType = typeof(InvoiceApply).FullName;
                    item.AddOnlyProperties(o => o.BizType);
                    item.BizKvid = newInvoiceApplyKvid;
                    item.AddOnlyProperties(o => o.BizKvid);
                    item.Status = (int)InvoiceApplyStatus.ApplyMerge;
                    item.AddOnlyProperties(o => o.Status);
                    conn.UpdateOnly(item);
                }
                #endregion

                #region 对应收款关联的ApplyKvid变更
                if (!relationApplys.IsNullOrEmpty())
                {
                    foreach (var relation in relationApplys)
                    {
                        relation.ParentKvid = newInvoiceApplyKvid;
                        relation.AddOnlyProperties(o => o.ParentKvid);
                        relation.RootKvid = newInvoiceApplyKvid;
                        relation.AddOnlyProperties(o => o.RootKvid);
                        conn.UpdateOnly(relation);
                    }
                }
                #endregion

                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return rtns;
        }
    }

    [Api(Description = "修改申请类型,专票普票申请之间的改变")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class InvoiceApplyChangeType : RestfulExecution<InvoiceApply>
    {
        public Guid Kvid { get; set; }

        public List<Guid> Kvids { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Kvid == Guid.Empty && Kvids.IsNullOrEmpty()).ThrowIfTrue("Need Kvid Or Kvids!");
            if (Kvids.IsNullOrEmpty()) Kvids = new List<Guid>();
            if (Kvid != null) Kvids.Add(Kvid);

            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var applys = conn.SelectByIds<InvoiceApply>(Kvids);
            applys.ThrowIfNullOrEmpty("未找到对应的申请!");
            (applys.Exists(o => o.Status > (int)InvoiceApplyStatus.FinancialExecute)).ThrowIfTrue("申请已执行开票,不可改变");

            var rtns = new RestfulUpdateResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            foreach (var item in applys)
            {
                if (item.Type == "VAT")
                {
                    item.Type = "VATS";
                    item.AddOnlyProperties(o => o.Type);
                    conn.UpdateOnly(item);
                    item.RemoveAllOnlyProperties();
                    rtns.Results.Add(item);
                }
                else if (item.Type == "VATS")
                {
                    item.Type = "VAT";
                    item.AddOnlyProperties(o => o.Type);
                    conn.UpdateOnly(item);
                    item.RemoveAllOnlyProperties();
                    rtns.Results.Add(item);
                }
            }

            return rtns;
        }
    }
    #endregion

    #region 开票客户端对接
    [RequiresAnyRole(SystemRoles.Everyone)]
    [Api(Description = "获取一条")]
    public class InvoiceApplyExecuting : RestfulExecution<InvoiceApply>
    {
        /// <summary>
        /// 优先级OperatorKvid>OwnerKvid>OrganizationKvid
        /// </summary>
        public Guid OperatorKvid { get; set; }
        public Guid OwnerKvid { get; set; }
        public Guid OrganizationKvid { get; set; }
        public override object OnExecution(IRequest req, IResponse res)
        {
            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var query = conn.From<InvoiceApply>();
            query.Where(o => o.Status == (int)InvoiceApplyStatus.FinancialExecute);
            if (OperatorKvid != Guid.Empty) query.And(o => o.OperatorKvid == OperatorKvid);
            else if (OwnerKvid != Guid.Empty) query.And(o => o.OwnerKvid == OwnerKvid);
            else if (OrganizationKvid != Guid.Empty) query.And(o => o.OrganizationKvid == OrganizationKvid);
            var apply = conn.Single(query);

            var rtns = new RestfulQueryResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            if (apply == null) return rtns;

            apply.Status = (int)InvoiceApplyStatus.AutoExecuting;
            apply.AddOnlyProperties(o => o.Status);
            conn.UpdateOnly(apply);
            apply.RemoveAllOnlyProperties();
            rtns.Results.Add(apply);

            rtns.Total = rtns.Results.Count;
            return rtns;
        }
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    [Api(Description = "开票结果回写")]
    public class InvoiceApplyResult : RestfulExecution<InvoiceApply>
    {
        public Guid ApplyKvid { get; set; }

        public string FileContent { get; set; }

        //发票开具最大额度限制 默认10万
        public decimal Limit { get; set; } = 100000;

        public override object OnExecution(IRequest req, IResponse res)
        {
            ApplyKvid.ThrowIfEmpty("发票申请Kvid 不能为空!");
            FileContent.ThrowIfNullOrEmpty("开票结果内容不能为空!");
            var contents = FileContent.Replace("\r", "\n").Split('\n').ToList();
            List<List<string>> infos = new List<List<string>>();
            foreach (var content in contents)
            {
                if (content.IsNullOrEmpty()) continue;
                var info = content.Replace(" ", "").Replace("[", "").Replace("]", ",").Replace("\n", "").Split(',').ToList();
                (info.Count() < 3).ThrowIfTrue("文件内容格式不正确!");
                infos.Add(info);
            }
            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var apply = conn.SingleById<InvoiceApply>(ApplyKvid);
            apply.ThrowIfNull("未找到对应发票申请信息!");
            (apply.AmountInvoice == apply.Amount).ThrowIfTrue("此发票申请已开具发票!");

            var applyDetails = conn.Select<InvoiceApplyDetail>(o => o.ApplyKvid == ApplyKvid);
            applyDetails.ThrowIfNullOrEmpty("未找到对应发票申请明细信息!");
            //拆分applyDetail (超10万的)
            var applyDetailSplited = new List<InvoiceApplyDetail>();
            //只有一条明细的情况下 需要对明细进行拆分超过10万的部分
            if (applyDetails.Count == 1) applyDetailSplited = applyDetails[0].SplitApplyDetail(Limit);
            //多条明细的情况 不进行拆分了  但要判断下要是超过10万就要报错了
            else
            {
                var group = applyDetails.GroupBy(o => o.GoodsId);
                foreach (var kv in group)
                {
                    var sum = kv.Sum(o => o.Amount);
                    if (sum > Limit) throw new Exception($"存在单张发票明细超过{Limit}元 无法创建!");
                }
                applyDetailSplited = applyDetails;
            }
            (applyDetailSplited.Sum(o => o.Amount) != apply.Amount).ThrowIfTrue("要生成的发票金额之和不等于申请金额!");
            var applyRelations = conn.Select<InvoiceApply>(o => o.ParentKvid == apply.Kvid && o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty);

            var logs = new List<EntityLog<InvoiceApply>>();
            var updateApplys = new List<InvoiceApply>();
            var insertInvoices = new List<Invoice>();
            var insertInvoiceDetails = new List<InvoiceDetail>();

            foreach (var info in infos)
            {
                #region 开票失败的情况处理
                if (info[2] != "开具结果:1")
                {
                    apply.Status = (int)InvoiceApplyStatus.ErrorInsertInvoice;
                    apply.AddOnlyProperties(o => o.Status);
                    apply.Remark = $"开票失败! {info.Join(",")}";
                    apply.AddOnlyProperties(o => o.Remark);
                    //apply.AmountUsed = 0;
                    //apply.AddOnlyProperties(o => o.AmountUsed);
                    updateApplys.Add(apply);
                    //conn.UpdateOnly(apply);

                    if (!applyRelations.IsNullOrEmpty())
                    {
                        foreach (var item in applyRelations)
                        {
                            item.Status = -1;
                            item.AddOnlyProperties(o => o.Status);
                            updateApplys.Add(item);
                            //conn.UpdateOnly(item);
                        }
                    }

                    var log = new EntityLog<InvoiceApply>();
                    log.OwnerKvid = apply.Kvid;
                    log.BizId = apply.SerialNumber;
                    log.BizKvid = apply.Kvid;
                    log.BizType = typeof(InvoiceApply).FullName;
                    log.Title = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                    log.Remark = $"编号:{apply.SerialNumber} 申请失败,原因:{info.Join(",")}";
                    logs.Add(log);
                    continue;
                }
                #endregion

                #region 开票成功的情况处理
                var time = DateTime.ParseExact(info[0], "yyyy-MM-ddHH:mm:ss", CultureInfo.InvariantCulture);//开票时间
                var code = info[4];//开票代码
                var number = System.Text.RegularExpressions.Regex.Replace(info[5], @"[^0-9]+", "");//开票号码,唯一
                var type = info[3].Split(':')[1];//开票类型
                var group = info[1].Split(':')[1];
                var currentDetails = group.IsNullOrEmpty() ? applyDetailSplited : applyDetailSplited.Where(o => o.GoodsId == group).ToList();

                var invoiceKvid = Guid.NewGuid();
                var invoice = new Invoice();

                invoice.RootKvid = invoiceKvid;
                invoice.Kvid = invoiceKvid;
                invoice.ApplyKvid = ApplyKvid;
                invoice.OwnerKvid = apply.OwnerKvid;
                invoice.OwnerName = apply.OwnerName;
                invoice.Currency = CurrencyUnit.CNY;
                invoice.Type = type;
                invoice.Category = apply.OperateType.ToString();
                invoice.OperateTime = time;
                invoice.SerialCode = code;
                invoice.SerialNumber = number;
                invoice.Status = 1;
                invoice.Amount = currentDetails.Sum(o => o.Amount);
                invoice.AmountTax = currentDetails.Sum(o => o.AmountTax);
                invoice.AmountUntaxed = currentDetails.Sum(o => o.AmountUntaxed);
                invoice.TaxRate = currentDetails[0].TaxRate;
                invoice.PayeeKvid = apply.PayeeKvid;
                invoice.PayeeName = apply.PayeeName;
                invoice.PayeeTaxNumber = apply.PayeeTaxNumber;
                invoice.PayeeCompanyAddress = apply.PayeeCompanyAddress;
                invoice.PayeePhone = apply.PayeePhone;
                invoice.PayeeRegisteredBank = apply.PayeeRegisteredBank;
                invoice.PayeeBankAccount = apply.PayeeBankAccount;
                invoice.PayerKvid = apply.PayerKvid;
                invoice.PayerName = apply.PayerName;
                invoice.PayerTaxNumber = apply.PayerTaxNumber;
                invoice.PayerCompanyAddress = apply.PayerCompanyAddress;
                invoice.PayerPhone = apply.PayerPhone;
                invoice.PayerRegisteredBank = apply.PayerRegisteredBank;
                invoice.PayerBankAccount = apply.PayerBankAccount;
                invoice.OperatorName = apply.OperatorName;
                invoice.Remark = apply.Remark;

                invoice.Metadata = new Dictionary<string, string>();
                invoice.Metadata["PayeeOperatorName"] = apply.Metadata.ContainsKey("PayeeOperatorName") ? apply.Metadata["PayeeOperatorName"] : "";
                invoice.Metadata["ReviewerName"] = apply.Metadata.ContainsKey("ReviewerName") ? apply.Metadata["ReviewerName"] : "";
                insertInvoices.Add(invoice);
                //conn.Insert(invoice);

                foreach (var item in currentDetails)
                {
                    var detail = new InvoiceDetail();
                    detail.InvoiceKvid = invoiceKvid;
                    detail.GoodsFullName = item.GoodsFullName;
                    detail.GoodsSpecifications = item.GoodsSpecifications;
                    detail.Quantity = item.Quantity;
                    detail.QuantityUnitPriceUntaxed = item.QuantityUnitPriceUntaxed;
                    detail.AmountUntaxed = item.AmountUntaxed;
                    detail.Amount = item.Amount;
                    detail.AmountTax = item.AmountTax;
                    detail.TaxRate = item.TaxRate;
                    insertInvoiceDetails.Add(detail);
                    //conn.Insert(detail);
                }
                #endregion

                #region 更新申请状态
                if (!applyRelations.IsNullOrEmpty())
                {
                    foreach (var item in applyRelations)
                    {
                        //无需更新
                        if (item.AmountInvoice >= item.Amount) continue;
                        var relationAmount = item.Amount - item.AmountInvoice;
                        var amount = relationAmount <= invoice.Amount ? relationAmount : invoice.Amount;
                        item.AmountInvoice += amount;
                        item.AddOnlyProperties(o => o.AmountInvoice);
                        if (item.AmountInvoice == item.Amount)
                        {
                            item.Status = (int)InvoiceApplyStatus.Completed;
                            item.AddOnlyProperties(o => o.Status);
                        }
                        updateApplys.Add(item);
                        //conn.UpdateOnly(item);
                    }
                }
                apply.AmountInvoice += invoice.Amount;
                apply.AddOnlyProperties(o => o.AmountInvoice);
                if (apply.AmountInvoice == apply.Amount)
                {
                    apply.Status = (int)InvoiceApplyStatus.Completed;
                    apply.AddOnlyProperties(o => o.Status);
                }
                updateApplys.Add(apply);
                //conn.UpdateOnly(apply);
                #endregion
            }

            #region 收款与发票关联
            var payments = conn.Select<Payment>(o => Sql.In(o.Kvid, applyRelations.ConvertAll(p => p.BizKvid)));//Type:Bank,Split,Pos,Cash,WeChat...
            List<Payment> parentPayments = new List<Payment>();//也要更新父级的AmountInvoice
            if (!payments.IsNullOrEmpty())
            {
                var correlatedInvoices = insertInvoices.Correlating(payments);
                insertInvoices.AddRange(correlatedInvoices);
                var currentRootKvids = payments.Where(o => o.Type==PaymentType.Split).ToList().ConvertAll(p => p.RootKvid);
                if (!currentRootKvids.IsNullOrEmpty()) parentPayments = conn.Select<Payment>(o => o.OffsetKvid == Guid.Empty && Sql.In(o.Kvid, currentRootKvids));
            }
            #endregion

            var rtns = new RestfulUpdateResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var trans = conn.OpenTransaction();

            try
            {
                if (!logs.IsNullOrEmpty()) logs.ForEach(o => conn.Insert(o));
                if (!updateApplys.IsNullOrEmpty())
                {
                    updateApplys.ForEach(o => conn.UpdateOnly(o));
                    rtns.Results.AddRange(updateApplys);
                }
                if (!insertInvoices.IsNullOrEmpty())
                {
                    foreach (var item in insertInvoices)
                    {
                        if (item.Type == "Relation") item.SerialNumber = item.GetSubSerialNumber(conn);
                        conn.Insert(item);
                    }
                }
                if (!insertInvoiceDetails.IsNullOrEmpty()) insertInvoiceDetails.ForEach(o => conn.Insert(o));

                if (!payments.IsNullOrEmpty())
                {
                    payments.ForEach(o => conn.UpdateOnly(o));
                }
                if (!parentPayments.IsNullOrEmpty())
                {
                    parentPayments.ForEach(o => o.RecalculateAmountInvoice(conn));
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }

            return rtns;
        }
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    [Api(Description = "开票结果回写2")]
    public class InvoiceApplyResult2 : RestfulExecution<InvoiceApply>
    {
        public Guid ApplyKvid { get; set; }

        public List<Invoice> Items { get; set; }

        //发票开具最大额度限制 默认10万
        public decimal Limit { get; set; } = 100000;

        public override object OnExecution(IRequest req, IResponse res)
        {
            ApplyKvid.ThrowIfEmpty("发票申请Kvid 不能为空!");
            Items.ThrowIfNullOrEmpty("开票结果内容不能为空!");
            foreach (var item in Items)
            {
                (item.OperateTime == DateTime.MinValue || item.OperateTime == DateTime.MaxValue).ThrowIfTrue("缺少开票日期信息!");
                item.SerialCode.ThrowIfNullOrEmpty("缺少发票代码信息!");
                item.SerialNumber.ThrowIfNullOrEmpty("缺少发票号码信息!");
                item.Type.ThrowIfNullOrEmpty("缺少发票类型信息!");
                (item.TaxRate < 0 || item.TaxRate > 1).ThrowIfTrue("缺少发票税率信息!");
                (item.Amount == 0).ThrowIfTrue("缺少发票金额信息!");
            }
            var conn = KiviiContext.GetOpenedDbConnection<InvoiceApply>();
            var apply = conn.SingleById<InvoiceApply>(ApplyKvid);
            apply.ThrowIfNull("未找到对应发票申请信息!");
            (apply.AmountInvoice == apply.Amount).ThrowIfTrue("此发票申请已开具发票!");
            var applyRelations = conn.Select<InvoiceApply>(o => o.ParentKvid == apply.Kvid && o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty);

            var logs = new List<EntityLog<InvoiceApply>>();
            var updateApplys = new List<InvoiceApply>();
            var insertInvoices = new List<Invoice>();
            var insertInvoiceDetails = new List<InvoiceDetail>();

            foreach (var info in Items)
            {
                #region 开票成功的情况处理
                var time = info.OperateTime;//DateTime.ParseExact(info[0], "yyyy-MM-ddHH:mm:ss", CultureInfo.InvariantCulture);//开票时间
                var code = info.SerialCode;//info[4];//开票代码
                var number = info.SerialNumber;//System.Text.RegularExpressions.Regex.Replace(info[5], @"[^0-9]+", "");//开票号码,唯一
                var type = info.Type;//info[3].Split(':')[1];//开票类型
                var currentAmount = info.Amount;
                var taxRate = info.TaxRate;

                var invoiceKvid = Guid.NewGuid();
                var invoice = new Invoice();

                invoice.RootKvid = invoiceKvid;
                invoice.Kvid = invoiceKvid;
                invoice.ApplyKvid = ApplyKvid;
                invoice.OwnerKvid = apply.OwnerKvid;
                invoice.OwnerName = apply.OwnerName;
                invoice.Currency = CurrencyUnit.CNY;
                invoice.Type = type;
                invoice.Category = apply.OperateType.ToString();
                invoice.OperateTime = time;
                invoice.SerialCode = code;
                invoice.SerialNumber = number;
                invoice.Status = 1;
                invoice.Amount = currentAmount;
                invoice.AmountUntaxed = Math.Round(currentAmount / (1 + taxRate), 2);
                invoice.AmountTax = currentAmount - invoice.AmountUntaxed;
                invoice.TaxRate = taxRate;
                invoice.PayeeKvid = apply.PayeeKvid;
                invoice.PayeeName = apply.PayeeName;
                invoice.PayeeTaxNumber = apply.PayeeTaxNumber;
                invoice.PayeeCompanyAddress = apply.PayeeCompanyAddress;
                invoice.PayeePhone = apply.PayeePhone;
                invoice.PayeeRegisteredBank = apply.PayeeRegisteredBank;
                invoice.PayeeBankAccount = apply.PayeeBankAccount;
                invoice.PayerKvid = apply.PayerKvid;
                invoice.PayerName = apply.PayerName;
                invoice.PayerTaxNumber = apply.PayerTaxNumber;
                invoice.PayerCompanyAddress = apply.PayerCompanyAddress;
                invoice.PayerPhone = apply.PayerPhone;
                invoice.PayerRegisteredBank = apply.PayerRegisteredBank;
                invoice.PayerBankAccount = apply.PayerBankAccount;
                invoice.OperatorName = apply.OperatorName;
                invoice.Remark = apply.Remark;

                invoice.Metadata = new Dictionary<string, string>();
                invoice.Metadata["PayeeOperatorName"] = apply.Metadata.ContainsKey("PayeeOperatorName") ? apply.Metadata["PayeeOperatorName"] : "";
                invoice.Metadata["ReviewerName"] = apply.Metadata.ContainsKey("ReviewerName") ? apply.Metadata["ReviewerName"] : "";
                insertInvoices.Add(invoice);
                //conn.Insert(invoice);

                var detail = new InvoiceDetail();
                detail.InvoiceKvid = invoiceKvid;
                detail.GoodsFullName = "录入明细";//item.GoodsFullName;
                                              //detail.GoodsSpecifications = item.GoodsSpecifications;
                detail.Quantity = 1;//item.Quantity;
                detail.AmountUntaxed = invoice.AmountUntaxed;
                detail.Amount = invoice.Amount;
                detail.AmountTax = invoice.AmountTax;
                detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 2);
                detail.TaxRate = invoice.TaxRate;
                insertInvoiceDetails.Add(detail);
                //conn.Insert(detail);
                #endregion

                #region 更新申请状态
                if (!applyRelations.IsNullOrEmpty())
                {
                    foreach (var item in applyRelations)
                    {
                        //无需更新
                        if (item.AmountInvoice >= item.Amount) continue;
                        var relationAmount = item.Amount - item.AmountInvoice;
                        var amount = relationAmount <= invoice.Amount ? relationAmount : invoice.Amount;
                        item.AmountInvoice += amount;
                        item.AddOnlyProperties(o => o.AmountInvoice);
                        if (item.AmountInvoice == item.Amount)
                        {
                            item.Status = (int)InvoiceApplyStatus.Completed;
                            item.AddOnlyProperties(o => o.Status);
                        }
                        updateApplys.Add(item);
                        //conn.UpdateOnly(item);
                    }
                }
                apply.AmountInvoice += invoice.Amount;
                apply.AddOnlyProperties(o => o.AmountInvoice);
                if (apply.AmountInvoice == apply.Amount)
                {
                    apply.Status = (int)InvoiceApplyStatus.Completed;
                    apply.AddOnlyProperties(o => o.Status);
                }
                updateApplys.Add(apply);
                //conn.UpdateOnly(apply);
                #endregion
            }

            #region 收款与发票关联
            var payments = conn.Select<Payment>(o => Sql.In(o.Kvid, applyRelations.ConvertAll(p => p.BizKvid)));//Type:Bank,Split,Pos,Cash,WeChat...
            List<Payment> parentPayments = new List<Payment>();//也要更新父级的AmountInvoice
            if (!payments.IsNullOrEmpty())
            {
                var correlatedInvoices = insertInvoices.Correlating(payments);
                insertInvoices.AddRange(correlatedInvoices);
                var currentRootKvids = payments.Where(o => o.Type == PaymentType.Split).ToList().ConvertAll(p => p.RootKvid);
                if (!currentRootKvids.IsNullOrEmpty()) parentPayments = conn.Select<Payment>(o => o.OffsetKvid == Guid.Empty && Sql.In(o.Kvid, currentRootKvids));
            }
            #endregion

            var rtns = new RestfulUpdateResponse<InvoiceApply>();
            rtns.Results = new List<InvoiceApply>();

            var trans = conn.OpenTransaction();

            try
            {
                if (!logs.IsNullOrEmpty()) logs.ForEach(o => conn.Insert(o));
                if (!updateApplys.IsNullOrEmpty())
                {
                    updateApplys.ForEach(o => conn.UpdateOnly(o));
                    rtns.Results.AddRange(updateApplys);
                }
                if (!insertInvoices.IsNullOrEmpty())
                {
                    foreach (var item in insertInvoices)
                    {
                        if (item.Type == "Relation") item.SerialNumber = item.GetSubSerialNumber(conn);
                        conn.Insert(item);
                    }
                }
                if (!insertInvoiceDetails.IsNullOrEmpty()) insertInvoiceDetails.ForEach(o => conn.Insert(o));

                if (!payments.IsNullOrEmpty())
                {
                    payments.ForEach(o => conn.UpdateOnly(o));
                }
                if (!parentPayments.IsNullOrEmpty())
                {
                    parentPayments.ForEach(o => o.RecalculateAmountInvoice(conn));
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }

            return rtns;
        }
    }
    #endregion
}
